#ifndef ENGINE_H
#define ENGINE_H

#include <allegro.h>
#include "object.h"
#include "button.h"
#include "settings.h"
#include "resources.h"
#include "player.h"
#include "level.h"
#include "monster.h"
#include <aldumb.h>
#include "fire.h"

using namespace std;

class Main;
class Player;
class PlayerState;

// struct Rect
// {
// 	public:
// 		int x, int y, int w, int h;
// 		Rect(int _x, int _y, int _w, int _h) : x(_x), y(_y), w(_w), h(_h) {}
// }

class Hiscores
{
		char name[10][3];
		int score[10];
		FONT *hiscorefont;
	public:
		
		void init(Resources *res); // restore from disk
		void done(); // save to disk
		bool isHiScore(int score);
		
// 		void initInputMode();
// 		void initShowMode();
		void draw(BITMAP *buffer);
// 		void update();
};

class Collection
{
		Objects objects;
		Monster *objectArray[MONSTER_NUM]; // weak references to Object's in objects
		static const int COLLECTION_AREA_TOP = 440;
		static const int COLLECTION_AREA_LEFT = 0;
		
		bool collection [MONSTER_NUM]; // collection means collected, after fighting monkey
		bool total [MONSTER_NUM]; // total means either available, on screen or in collection
		bool available [MONSTER_NUM]; // available meens not spawned yet
		int collectionCount;
		int availableLeft;
		int totalCount;
		Engine *parent;
	public:
		Collection (Engine *_parent);
		void update();
		void initLevel (int count);	
		int getNextAvailable();	// returns -1 on failure
		void addToCollection (int monkey);
		void draw (BITMAP *buffer);
		int getTotal () { return totalCount; }		
		bool hasCollectedAll() { return collectionCount == totalCount; }
};

class Particle
{
	public:
		fix x;
		fix y;
		fix dx;
		fix dy;
		fix r;
		fix fr;
		int life;
		bool alive;
};

class Smoke
{
	list<Particle> particles;
	
	public:
		void reset();
		void draw(BITMAP *buffer, int xofst, int yofst);
		void update();
		void add (int x, int y);
};

class Engine
{
public:
	enum GameState {GS_INTRO, GS_PLAY, GS_MENU_RESUME, GS_DONE, GS_OUTRO, GS_HISCORE };
private:
	
	static const int GAME_AREA_TOP = 80;
	static const int GAME_AREA_HEIGHT = 360;
	static const int GAME_AREA_WIDTH = 640;
	static const int GAME_AREA_LEFT = 0;
	static const int TRASH_SPRITE_NUM = 12;
	static const int POWER_MODE_TIME = 770;
	static const int INTRO_TIME = 80;
	static const int SPAWN_TIME = 200;
	static const int BANANA_TIME = 600;
	
	GameState state;
	GameState oldState;

	int introTimer;
	int outroTimer;
	int spawnTimer;

	unsigned int currentLevel; // index of current Level

	Main *parent;
	
	Objects objects;
	Player *player[2];
	PlayerState ps[2];
	bool debug;
	
	Settings *settings;
	Resources *resources;
	
	FONT *gamefont;	
	FONT *introfont;	
	string gameover_message;
	
	BITMAP *field;
	BITMAP *trash[TRASH_SPRITE_NUM];
	RLE_SPRITE *skyline;
	BITMAP *cnn;
	BITMAP *playerSprite[2];
	int powerMode;
	Fire fire;
	void stopPowerMode();

public:
	LevelState ls;
	
	Engine (Main *p);
	Resources *getResources() { return resources; }
	Main *getParent() { return parent; }
	void initGame ();
	void init(); // call once during startup
	void initBg(int r, int b, int g);
	void drawTunnel (const std::string &s);
	void resume ();
	void draw (BITMAP *buffer);
	void update ();
	void spawn();
	
	GameState getState () { return state; }
	
	void nextLevel(); // called when player reaches exit
	void initLevel (); // for beginning a new level
	
	void doneLevel (); // clean up objects etc.
	void done(); // stop music
	bool isDebug () { return debug; }
	int getCurrentLevel () { return currentLevel; }

	int getCounter ();
		
	void playSample (SAMPLE *s);
	void playMusic (const char *id, float volume = 1.0f);
	void stopMusic ();
	Player *getNearestPlayer (int x, int y);
	
	Button* getButton(int p);
	
	Collection collection;
	Smoke smoke;
	
	void drawPlayerStats(BITMAP *buffer);
	
	bool getPowerMode() { return powerMode > 0; }
	void setPowerMode();
	AL_DUH_PLAYER *dp;
};
#endif
