#ifndef OBJECT_H
#define OBJECT_H

#include <allegro.h>
#include <list>
#include <map>

class Objects;
class Engine;
class Anim;
class Resources;

#define UP 0
#define DOWN 1
#define LEFT 2
#define RIGHT 3

enum ObjectType { OT_MONSTER, OT_BULLET, OT_PLAYER, OT_BONUS, OT_POWER_BONUS, OT_TIRE };

static const int dir_mult[2][4] = { { 0, 0, -1, 1 }, {-1, 1, 0, 0 } };
static const int opposite[4] = { DOWN, UP, RIGHT, LEFT };

// in this class,
// current is not guaranteed to be set.
// parent is not guaranteed to be set,
// although it should be when the first update / draw is called.
class Room;

class Object
{
private:
	int frame;
	int animcounter;	

	fix x, y;
	Anim *current;
	
	Resources *getResources(); // depracated
	int dir;
	
protected:

	BITMAP *field; // warning: field may be null
		
	int sprdx, sprdy;
	int w, h;
	bool solid;
	bool alive;
	bool visible;
	void setAnim (Anim *a);
	Objects* parent;
	Engine *e;
	
	void kill () { alive = false; }
public:
	int getDir() { return dir; }
	void setDir(int _dir);
	
	Object (BITMAP *_field);
	virtual ~Object () {}
	void setParent (Objects *p);
	
	bool isSolid() { return solid; } // can collide with other stuff?
	bool isVisible() { return visible; } // to be drawn?
	bool isAlive() { return alive; } // if not, scheduled to be killed
			
	RLE_SPRITE *getSprite ();
	
	void try_move (fix dx, fix dy);
	virtual void setLocation (fix nx, fix ny);
	int getSprx () { return x + sprdx; }
	int getSpry () { return y + sprdy; }
	fix getx () { return x; }
	fix gety () { return y; }
	int getw () { return w; }
	int geth () { return h; }
	virtual void update();
	virtual ObjectType getType () = 0;
	virtual void handleCollission(Object *o) = 0;
};

class Objects
{
private:
	Engine *parent;
	std::list <Object*> objects;	
public:
	Objects (Engine *p) { parent = p; }
	Engine *getParent() { return parent; }
	void add(Object *o);
	
	void update();
	void killAll();
	void draw (BITMAP *buffer, int cx, int cy, int cw, int ch, int xofst, int yofst);
	int size() { return objects.size(); }; // returns # of objects
};

#endif
