#ifndef PLAYER_H
#define PLAYER_H

#include "object.h"
#include "settings.h"

class Resources;
class Player;
	
class Bullet : public Object
{
	private:
		
		fix dx, dy;
		
		Player *parent;
		static Anim *bulletSprite;
	public:
		Player *getPlayer() { return parent; }
		Bullet(BITMAP *_field, int _dir, Player *_parent);
		
		virtual void update();
		virtual ObjectType getType () { return OT_BULLET; }
		virtual void handleCollission(Object *o);
		static void init(Resources *res);		
};

class Bonus : public Object
{
	private:
		static Anim *sprite;
	public:
		Bonus(BITMAP *_field, int x, int y);
		
		virtual ObjectType getType () { return OT_BONUS; }
		virtual void handleCollission(Object *o);
		static void init(Resources *res);		
};

class PowerBonus : public Bonus
{
	private:
		static Anim *sprite;
	public:
		PowerBonus(BITMAP *_field, int x, int y);
		virtual ObjectType getType () { return OT_POWER_BONUS; }
		static void init(Resources *res);		
};

class Tire : public Object
{
	private:
		fix dx, dy;		
		static Anim *sprite;
	public:
		Tire(BITMAP *_field, int x, int y);
		
		virtual void update();
		virtual ObjectType getType () { return OT_TIRE; }
		virtual void handleCollission(Object *o);
		static void init(Resources *res);		
};

class PlayerState
{
	public:
	int score;
	int life;
	int defaultx;
	int defaulty;
	
	PlayerState();
};

class Player : public Object
{
	static const int transportDelay = 5;
	static const int SMOKE_DELAY = 8;
	static const int SCORE_MONSTER = 125;
	static const int SCORE_BONUS = 25;
	static const int SCORE_OCTET = 250;
	static const int OCTET_TIME = 100;
	
	int transportCounter;

	static Anim *walk[2];
	
	int hittimer;
	int attacktimer; // period between shots
	int smoketimer; // period between smoke particles
	
	int octettimer; // see if we can get 8 within OCTET_TIME
	int octetcount;
	
	bool isWalking;
		
	int playerType; // 0 for fole, 1 for raul
	
	Button *button;
	
	void die();
public:
	PlayerState *ps;
	Player(BITMAP *_field, PlayerState *ps, int _playerType);
	virtual void update();
	static void init(Resources *res);
	virtual ObjectType getType () { return OT_PLAYER; }
	virtual void handleCollission(Object *o);
	bool canAttack() { return attacktimer == 0; }
	void setDefaultLocation();
};

#endif
