#include <assert.h>
#include "engine.h"
#include "main.h"
#include "object.h"
#include <stdio.h>
#include "player.h"
#include "colors.h"
#include "screenshot.h"

int Engine::getCounter () 
{ 
	return parent->getCounter();
}

Engine::Engine (Main *p) : objects (this), collection(this)
{
	parent = p;
	player[0] = NULL;
	player[1] = NULL;
	debug = false;
	resources = p->getResources();
	settings = p->getSettings();
	dp = NULL;
}

void Engine::doneLevel()
{
	objects.killAll();
}

Button* Engine::getButton(int p)
{ 
	return settings->button[p]; 
}

void Engine::init()
{
 	Player::init(resources);	
 	Bullet::init(resources);
 	Monster::init(resources);
 	Bonus::init(resources);
 	PowerBonus::init(resources);
 	Tire::init(resources);
 	fire.init();
 	
	gamefont = resources->getFont("tabletron_16");
	introfont = resources->getFont("tabletron_24");
 	srand(time(0));	
 	
 	field = create_bitmap (GAME_AREA_WIDTH, GAME_AREA_HEIGHT);
 	
 	// aggregate the various trash sprites
 	trash[0] = resources->getBitmap ("COMPU");
 	trash[1] = resources->getBitmap ("CHOCO");
 	trash[2] = resources->getBitmap ("CDS");
 	trash[3] = resources->getBitmap ("CHEESE");
 	trash[4] = resources->getBitmap ("GUN");
 	trash[5] = resources->getBitmap ("TIRE");
 	trash[6] = resources->getBitmap ("VISTA");
 	trash[7] = resources->getBitmap ("TOM");
 	trash[8] = resources->getBitmap ("BONO");
 	trash[9] = resources->getBitmap ("LEAF");
 	trash[10] = resources->getBitmap ("VCR");
 	trash[11] = resources->getBitmap ("CELL");

	skyline = resources->getRle ("skyline");
	cnn = resources->getBitmap ("cnn");
	
	RLE_SPRITE *temp = resources->getRle ("folebull-side4");
	playerSprite[0] = create_bitmap (temp->w, temp->h); // TODO: destroy_bitmap never called
	clear_to_color (playerSprite[0], MAGIC_PINK);
	draw_rle_sprite (playerSprite[0], temp, 0, 0);
	
	temp = resources->getRle ("raulbull-side2");
	playerSprite[1] = create_bitmap (temp->w, temp->h); // TODO: destroy_bitmap never called
	clear_to_color (playerSprite[1], MAGIC_PINK);
	draw_rle_sprite (playerSprite[1], temp, 0, 0);
}

void Engine::initGame ()
{
	// initialize game and player stats.
	// to start a new game.
	currentLevel = 0;

	ps[0] = PlayerState();
	ps[0].defaultx = 8;
	ps[0].defaulty = 9;
	ps[1] = PlayerState();
	ps[1].defaultx = 1;
	ps[1].defaulty = 1;
	
	initLevel();
}

void Engine::initLevel()
{
	// initialize objects
	
	if (ps[0].life > 0)
	{
		player[0] = new Player (field, &ps[0], 0);
		player[0]->setDefaultLocation();	

		objects.add (player[0]);
	}
	if (settings->numPlayers == 2 && ps[1].life > 0)
	{
		player[1] = new Player (field, &ps[1], 1);
		player[1]->setDefaultLocation();
		
		objects.add (player[1]);		
	}
	else
	{
		player[1] = NULL;
	}
		
	state = GS_INTRO; // show cutscene
	introTimer = INTRO_TIME;
	
	switch (currentLevel % 3)
	{
		case 0:	
			// initialize bitmaps
			initBg(80, 120, 40);
			
			drawTunnel ("1 1 S 3");
			drawTunnel ("1 5 E 4");
			drawTunnel ("5 5 S 4");
			drawTunnel ("19 1 W 4");
			drawTunnel ("14 2 S 7");
			drawTunnel ("14 9 W 7");
			objects.add (new Tire (field, 3,3));
			objects.add (new Tire (field, 12,3));
			objects.add (new Tire (field, 7,2));
			objects.add (new Tire (field, 16,3));
						
			objects.add (new Bonus (field, 1,8));
			objects.add (new Bonus (field, 1,9));
			objects.add (new Bonus (field, 1,10));
			objects.add (new Bonus (field, 2,8));
			objects.add (new Bonus (field, 2,9));
			objects.add (new Bonus (field, 2,10));
			objects.add (new Bonus (field, 3,8));
			objects.add (new Bonus (field, 3,9));
			objects.add (new Bonus (field, 3,10));
			
			objects.add (new Bonus (field, 19,6));
			objects.add (new Bonus (field, 19,7));
			objects.add (new Bonus (field, 18,6));
			objects.add (new Bonus (field, 18,7));
			
			objects.add (new Bonus (field, 5,3));
			objects.add (new Bonus (field, 6,3));
			objects.add (new Bonus (field, 7,3));
			objects.add (new Bonus (field, 8,3));
			objects.add (new Bonus (field, 9,3));
			objects.add (new Bonus (field, 10,3));
				
			break;
		case 1:
			// initialize bitmaps
			initBg(40, 80, 120);
			
			drawTunnel ("1 1 S 3");
			drawTunnel ("1 5 E 4");
			drawTunnel ("5 5 S 4");
			drawTunnel ("19 1 W 4");
			drawTunnel ("14 2 S 7");
			drawTunnel ("14 9 W 7");
			
			objects.add (new Tire (field, 3,3));
			objects.add (new Tire (field, 12,3));
			objects.add (new Tire (field, 7,2));
			objects.add (new Tire (field, 16,3));
						
			objects.add (new Bonus (field, 8,5));
			objects.add (new Bonus (field, 8,6));
			objects.add (new Bonus (field, 8,7));
			objects.add (new Bonus (field, 9,6));
			objects.add (new Bonus (field, 9,5));
			objects.add (new Bonus (field, 9,7));
			objects.add (new Bonus (field, 10,5));
			objects.add (new Bonus (field, 10,6));
			objects.add (new Bonus (field, 10,7));
			
			objects.add (new Bonus (field, 19,9));
			objects.add (new Bonus (field, 19,10));
			objects.add (new Bonus (field, 18,10));
			
			objects.add (new Bonus (field, 0,9));
			objects.add (new Bonus (field, 0,10));
			objects.add (new Bonus (field, 1,10));
			break;
		case 2:
			// initialize bitmaps
			initBg(120, 80, 40);
			
			drawTunnel ("1 1 S 3");
			drawTunnel ("1 5 E 4");
			drawTunnel ("5 5 S 4");
			drawTunnel ("19 1 W 4");
			drawTunnel ("14 2 S 7");
			drawTunnel ("14 9 W 7");
			
			objects.add (new Tire (field, 7,3));
			objects.add (new Tire (field, 10,3));
			
			objects.add (new Tire (field, 2,2));
			objects.add (new Tire (field, 18,5));
			objects.add (new Tire (field, 19,5));
						
			objects.add (new Bonus (field, 8,3));
			objects.add (new Bonus (field, 8,4));
			objects.add (new Bonus (field, 8,5));
			objects.add (new Bonus (field, 8,6));
			objects.add (new Bonus (field, 8,7));
			objects.add (new Bonus (field, 9,3));
			objects.add (new Bonus (field, 9,4));
			objects.add (new Bonus (field, 9,5));
			objects.add (new Bonus (field, 9,6));
			objects.add (new Bonus (field, 9,7));
			objects.add (new Bonus (field, 10,4));
			objects.add (new Bonus (field, 10,5));
			objects.add (new Bonus (field, 7,4));
			objects.add (new Bonus (field, 7,5));
			
			objects.add (new Bonus (field, 18,6));
			objects.add (new Bonus (field, 18,7));
			objects.add (new Bonus (field, 18,8));
			objects.add (new Bonus (field, 18,9));
			objects.add (new Bonus (field, 18,10));
			objects.add (new Bonus (field, 19,6));
			objects.add (new Bonus (field, 19,7));
			objects.add (new Bonus (field, 19,8));
			objects.add (new Bonus (field, 19,9));
			objects.add (new Bonus (field, 19,10));
			break;
	}
	 	
 	spawnTimer = 0;
 	
 	int mnum = currentLevel + 4;
 	if (mnum > 11) mnum = 11;
 	collection.initLevel(mnum);
 	
 	smoke.reset();
 	powerMode = 0;
}

void Engine::drawTunnel(const string &s)
{
	int x, y, l;
	char c;
	sscanf (s.c_str(), "%d %d %c %d", &x, &y, &c, &l);
	
	x *= 32;
	y *= 32;
	l++;
	
	int dx = 32;
	int dy = 32;
	
	switch (c)
	{
		case 'E': dx = l * 32;	break;
		case 'N': dy = l * -32;	break;
		case 'W': dx = l * -32;	break;
		case 'S': dy = l * 32;	break;
	}
	
	rectfill (field, x, y, x + dx, y + dy, MAGIC_PINK);
}

void Engine::initBg(int r, int g, int b)
{
	// draw lots of trash...
	
	BITMAP *temp = create_bitmap (GAME_AREA_WIDTH, GAME_AREA_HEIGHT);
	
	BITMAP *pattern = create_bitmap (32, 32);	
	for (int x = 0; x < 32; x += 2)
	{
		for (int y = 0; y < 32; y += 2)
		{
			int c = rand() % 128 + 64;
			rectfill (pattern, x, y, x+1, y+1, makecol (c, c, c));
		}
	}
		
	drawing_mode (DRAW_MODE_COPY_PATTERN, pattern, 0, 0);
	rectfill (temp, 0, 0, temp->w, temp->h, 0);
	drawing_mode (DRAW_MODE_SOLID, NULL, 0, 0);
	
	for (int x = 0; x < GAME_AREA_WIDTH; x += 30)
		for (int y = 0; y < GAME_AREA_HEIGHT; y += 30)
			{
				int dx = (rand() % 30) - 15;
				int dy = (rand() % 30) - 15;
				int r = rand() % 256;
				int spr = rand() % TRASH_SPRITE_NUM;
				int scale = 5 + rand() % 5;
				
				rotate_scaled_sprite (temp, trash[spr], x + dx, y + dy, 
					itofix(r), ftofix (scale / 10.0));
			}
	set_trans_blender (r, g, b, 128);
	draw_lit_sprite (field, temp, 0, 0, 180);
	
	destroy_bitmap (temp);
	destroy_bitmap (pattern);
}

void Engine::resume ()
{
	state = oldState;
}

// called when exit found
void Engine::nextLevel()
{
	// go on to next level
	stopPowerMode();
	state = GS_INTRO;
	introTimer = INTRO_TIME;
	currentLevel++;
	objects.killAll();
	initLevel();
}

void Engine::draw (BITMAP *buffer)
{
	switch (state)
	{
		case GS_PLAY:
			{
 				rectfill (buffer, 
 					GAME_AREA_LEFT, GAME_AREA_TOP, 
 					GAME_AREA_LEFT + GAME_AREA_WIDTH, GAME_AREA_TOP + GAME_AREA_HEIGHT, 
 					makecol (80, 60, 30));
 				if (getPowerMode())
 				{
 					fire.draw(buffer);
 				}
 				else
 				{
					rectfill (buffer, 
						0, 0, 
						buffer->w, GAME_AREA_TOP -1, 
						makecol (0, 0, 100));
				}
				rectfill (buffer, 
 					0, GAME_AREA_TOP + GAME_AREA_HEIGHT, 
 					buffer->w, buffer->h - GAME_AREA_TOP + GAME_AREA_HEIGHT,
 					BLACK);
				
				// draw stuff here
				masked_blit (field, buffer, 0, 0, 
					GAME_AREA_LEFT, GAME_AREA_TOP, 
					GAME_AREA_WIDTH, GAME_AREA_HEIGHT);
				draw_rle_sprite (buffer, skyline, 0, 0);
					
				smoke.draw (buffer, GAME_AREA_LEFT, GAME_AREA_TOP);
				objects.draw (buffer, 
					0, 0, 
					buffer->w, buffer->h, -GAME_AREA_LEFT, -GAME_AREA_TOP);
				collection.draw (buffer);
				drawPlayerStats(buffer);
				
			}
			break;
		case GS_OUTRO: 
			stretch_blit (cnn, buffer,
				0, 0, cnn->w, cnn->h,
				0, 0, buffer->w, buffer->h);
			textprintf_centre_ex (buffer, introfont, 
				340, 380, 
				RED, -1, 
				"%s", gameover_message.c_str());
			break;			
		case GS_INTRO:
			stretch_blit (cnn, buffer,
				0, 0, cnn->w, cnn->h,
				0, 0, buffer->w, buffer->h);
			textprintf_centre_ex (buffer, introfont, 
				340, 380, 
				RED, -1, 
				"LEVEL %i", currentLevel+1);
			break;
		case GS_DONE: case GS_MENU_RESUME:
			break;
		default:
			assert (false);
	}
}

void Engine::drawPlayerStats(BITMAP *buffer)
{
	int x = 90;
	for (int i = 0; i < ps[0].life-1; ++i)
	{
		masked_stretch_blit (playerSprite[0], buffer, 
			0, 0, playerSprite[0]->w, playerSprite[0]->h, 
			x, 0, 16, 16);
		x += 16;
	}
	textprintf_ex (buffer, gamefont, 
		0, 0, WHITE, -1, 
		"%06i", ps[0].score);
	textprintf_ex (buffer, gamefont, 
		0, 16, WHITE, -1, 
		"AMMO %i", player[0]->canAttack() ? 1 : 0);
	if (settings->numPlayers == 2)
	{
		x = buffer->w - 106;
		for (int i = 0; i < ps[1].life-1; ++i)
		{
			masked_stretch_blit (playerSprite[1], buffer, 
				0, 0, playerSprite[1]->w, playerSprite[1]->h, 
				x, 0, 16, 16);
			x -= 16;
		}
		textprintf_right_ex (buffer, gamefont, 
			buffer->w, 0, WHITE, -1, 
			"%06i", ps[1].score);
		textprintf_right_ex (buffer, gamefont, 
			buffer->w, 16, WHITE, -1, 
			"AMMO %i", player[1]->canAttack() ? 1 : 0);
	}
}

void Engine::spawn()
{
	spawnTimer ++;
	
	if (spawnTimer > SPAWN_TIME) // TODO get from level
	{
		spawnTimer = 0;
		// check collection		
		int next = collection.getNextAvailable();
		if (next >= 0)
		{
			Monster *monster = new Monster(field, next);
			objects.add (monster);
		}
		else
		{
			if (!getPowerMode())
			{
				objects.add (new PowerBonus (field, 17,1));
				spawnTimer = -BANANA_TIME;
			}
		}
	}
}

void Engine::update ()
{
	int c = 0;
	if (keypressed())
	{	
		c = readkey();
		if (c >> 8 == KEY_ESC)
		{
			oldState = state;
			state = GS_MENU_RESUME;
		}	
		if (c >> 8 == KEY_F12)
		{
			screenshot();
		}
#ifdef DEBUG
		if (c >> 8 == KEY_F11)
		{
			debug = !debug;
		}
#endif	
	}
	switch (state)
	{
		case GS_INTRO: // level intro screen showing
			introTimer--;
			if (introTimer == 0)
			{
				state = GS_PLAY;
				introTimer = INTRO_TIME;
			}
			break;
		case GS_OUTRO:
			outroTimer--;
			if (outroTimer == 0)
			{
				doneLevel();
				state = GS_DONE;	
				break;
			}
			break;
		case GS_PLAY: // game playing
		{
			if (powerMode > 0) 
			{ 
				powerMode--; 
				if (powerMode == 0)
				{
					stopPowerMode();
				}
			}
			
			// add objects based on position
			spawn();
			objects.update();
			collection.update();
			if (getPowerMode())
			{
				fire.update();
			}
			
			bool gameover = true;

			for (int i = 0; i < settings->numPlayers; ++i)
			{					
				if (ps[i].life > 0) gameover = false;
			}
			if (gameover)
			{
				gameover_message = "GAME OVER";
				outroTimer = 100;
				state = GS_OUTRO;
			}
			
			smoke.update();
			
			if (collection.hasCollectedAll())
			{
				nextLevel();
			}

		}
		break;
		case GS_MENU_RESUME: // do nothing, main will switch to resume menu.
			break;
		case GS_DONE: // do nothing, game is finished (won or lost)
			break;
		case GS_HISCORE: // ???
			break;
		default:
			assert (false); // shouldn't occur
	}  
}

void Engine::playSample (SAMPLE *s)
{
	if (!settings->isSoundOn()) return;
	play_sample (s, 127, 127, 1000, 0);
}

void Engine::playMusic (const char *id, float volume)
{
    if (!(settings->isSoundOn() && settings->isMusicOn())) return;
    if (dp)
    {
        al_stop_duh (dp);
        dp = NULL;
    }
    dp = al_start_duh (resources->getDuh (id), 2, 0, volume, 4096, 22050);
}

void Engine::stopMusic ()
{
    if (dp) 
    {
        al_stop_duh (dp);
        dp = NULL;
    }
}

void Engine::done()
{
	stopMusic();
}

void Collection::initLevel (int count)
{
	objects.killAll();
	assert (count <= MONSTER_NUM);
	for (int i = 0; i < MONSTER_NUM; ++i) 
	{ 
		total[i] = false; 
		collection[i] = false; 
		available[i] = false;
		objectArray[i] = NULL;
	}
	int x = COLLECTION_AREA_LEFT;
	int y = COLLECTION_AREA_TOP;
	for (int i = 0; i < count; ++i)
	{
		int r;
		do
		{
			r = rand() % MONSTER_NUM;
		} 
		while (total[r]);
		
		Monster *m = new Monster(NULL, r);
		m->setLocation ((fix)x, (fix)y);
		objects.add (m);
		objectArray[r] = m;
		x += 40;
		
		total[r] = true;
		available[r] = true;
	}
	collectionCount = 0;
	availableLeft = count;
	totalCount = count;
}

int Collection::getNextAvailable()	// returns -1 on failure
{
	if (availableLeft == 0) { return -1; }
	int r;
	do
	{
		r = rand() % MONSTER_NUM;
	}
	while (!available[r]);
	
	available[r] = false;
	availableLeft--;
	
	return r;
}

void Collection::addToCollection (int monkey) 
{
	assert (total[monkey]);
	assert (objectArray[monkey]);
	
	if (!collection[monkey]);
	{
		collection[monkey] = true;
		objectArray[monkey]->dummyAppear();
		collectionCount++;
	}	
}

void Collection::draw (BITMAP *buffer) 
{
	objects.draw(buffer, 0, 0, buffer->w, buffer->h, 0, 0);
}

void Collection::update()
{
	objects.update(); // update anims
}

void Smoke::reset()
{
	particles.clear();
}

void Smoke::draw(BITMAP *buffer, int xofst, int yofst)
{
	drawing_mode (DRAW_MODE_TRANS, NULL, 0, 0);
	for (list<Particle>::iterator i = particles.begin(); i != particles.end(); ++i)
	{
		if (i->alive)
		{
			set_trans_blender (0, 0, 0, 200 - (i->life * 2));
			circlefill (buffer, i->x + xofst, i->y + yofst, i->r, BLACK);
		}
	}
	drawing_mode (DRAW_MODE_SOLID, NULL, 0, 0);
}

class MyParticleRemover
{
   public:
	  bool operator()(Particle &o)
	  {
		 if (!o.alive)
		 {
			return 1;
		 }
		 return 0;
	  }
};

void Smoke::update()
{
	for (list<Particle>::iterator i = particles.begin(); i != particles.end(); ++i)
	{
		if (i->alive)
		{
			i->life++;
			if (i->life > 100) { i->alive = false; }
			i->y += i->dy;
			i->x += i->dx;
			i->r *= i->fr;
		}
	}
	
	// remove all that are not alive!
	particles.remove_if (MyParticleRemover());
}

void Smoke::add (int x, int y)
{
	Particle p;
	p.x = x;
	p.y = y;
	p.dy = (fix)-0.2;
	p.dx = fix (((rand() % 20) - 10) * 0.01);
	p.life = 0;
	p.alive = true;
	p.r = (fix)2;
	p.fr = (fix)1.015;
	particles.push_back (p);
}

Collection::Collection(Engine *_parent) : objects (_parent), parent (_parent)
{
}

void Hiscores::init(Resources *res)
{
	// load or defaults
	for (int i = 0; i < 10; ++i)
	{
		name[i][0] = 'M';
		name[i][1] = 'V';
		name[i][2] = 'I';
		score[i] = i * 100;
	}
	
	hiscorefont = res->getFont("tabletron_16");
}

void Hiscores::done()
{
	// save
}

bool Hiscores::isHiScore(int i)
{
	return (i > score[9]);
}

void Hiscores::draw (BITMAP *buffer)
{
	clear_to_color (buffer, BLACK);
	int x = 40;
	int y = 80;
	for (int i = 0; i < 10; ++i)
	{
		textprintf_ex (buffer, hiscorefont,
			x, y, WHITE, -1, "%c%c%c %i", name[i][0], name[i][1], name[i][2], score[i]
		);
	}
}

void Engine::setPowerMode()
{ 
	powerMode = POWER_MODE_TIME; 
	playMusic ("badlands1");
	fire.initPowerMode();
}

void Engine::stopPowerMode()
{ 
	powerMode = 0;
	playMusic ("f-castle");
}

Player *Engine::getNearestPlayer (int x, int y)
{
	if (settings->numPlayers == 1) return player[0];
	
	int dx1 = player[0]->getx() - x;
	int dy1 = player[0]->gety() - y;
	int dx2 = player[1]->getx() - x;
	int dy2 = player[1]->gety() - y;
	
	if (dx1 * dx1 + dy1 * dy1 < dx2 * dx2 + dy2 * dy2)
	{
		return player[0];
	}
	else
	{
		return player[1];
	}
}
