#include "monster.h"
#include "resources.h"
#include "colors.h"
#include <assert.h>
#include "engine.h"

Anim *Monster::monkey[MONSTER_NUM];
Anim *Monster::outline[MONSTER_NUM];
Anim *Monster::teleport[2];

Monster::Monster(BITMAP *_field, int _type) : Object (_field), 
	dx(0), dy(0), type(_type)
{
	setLocation ((fix)MONSTER_START_X * 32, (fix)MONSTER_START_Y * 32);
	if (field)
	{
		state = MS_MOVE;
 		setAnim (monkey[type]);
	}
	else
	{
		state = MS_DUMMY;
		setAnim (outline[type]);
	}
	visible = true;
	turntimer = 0;
}

bool Monster::isDirOk (int cx, int cy, int idir)
{
	int ix = cx + (dir_mult[0][idir] * getw() / 2);
	int iy = cy + (dir_mult[1][idir] * geth() / 2);
	int c = getpixel (field, ix, iy);
	return (c == MAGIC_PINK);
}

void Monster::update()
{
	timer ++;
	switch (state)
	{
	case MS_MOVE:
		{
			assert (field);
			
			// determine dir to nearest player
			Player *p = e->getNearestPlayer(getx(), gety());
			int pdx = p->getx() - getx();
			int pdy = p->gety() - gety();
			
			int mdir;
			if (abs (pdx) < abs(pdy))
			{
				if (pdx > 0) mdir = UP; else mdir = DOWN;
			}
			else
			{
				if (pdy > 0) mdir = LEFT; else mdir = RIGHT;
			}
			
			// chase or flee?
			if (e->getPowerMode())
			{
				mdir = opposite[mdir];
			}
			
			// direction check
			int cx = getx() + getw() / 2;
			int cy = gety() + geth() / 2;
			
			bool ok = false;
			// check current dir
			if (!ok && isDirOk(cx, cy, getDir()))
			{
				ok = true;
			}
			// check preferred dir
			if (!ok && mdir != opposite[getDir()] && isDirOk (cx, cy, mdir))
			{
				ok = true;
				setDir (mdir);
				turntimer = 0;
				break;
			}
			// check side dirs
			if (!ok)
			{
				for (int idir = 0; idir < 3; ++idir)
				{
					if (idir == opposite[getDir()]) { continue; }
					if (idir == getDir()) { continue; }
					if (isDirOk (cx, cy, idir))
					{
						ok = true;
						setDir (idir);
						turntimer = 0;
						break;
					}
				}
			}
			// lastly check opposite dir
			if (!ok)
			{
				if (isDirOk(cx, cy, opposite[getDir()]))
				{
					ok = true;
					setDir (opposite[getDir()]);
					turntimer = 0;
				}
			}
			// no dir is ok, we're stuck somehow. Don't move
			if (!ok)
			{
				dx = 0;
				dy = 0;
			}
			
			dx = fix(dir_mult[0][getDir()] * MONSTER_SPEED_FACTOR);
			dy = fix(dir_mult[1][getDir()] * MONSTER_SPEED_FACTOR);
			
			try_move (dx, dy);
		}
		break;
	case MS_APPEAR:
		if (timer >= TELEPORT_TIME)
		{
			setAnim (monkey[type]);
			state = MS_DUMMY;
		}
		break;
	case MS_DISAPPEAR:
		if (timer >= TELEPORT_TIME)
		{
			kill();
		}
		break;
	case MS_DUMMY:
		// nothing
		break;
	}
	
	// animate	
	Object::update();
}

void Monster::handleCollission(Object *o)
{
	if (state == MS_MOVE)
	{
		if (o->getType() == OT_PLAYER)
		{
			if (e->getPowerMode()) // power mode
			{
				die(); // disappear
			}
		}
		else if (o->getType() == OT_BULLET)
		{
			die(); // disappear
		}
	}
}

void Monster::die()
{
	state = MS_DISAPPEAR;
	setAnim (teleport[0]);
	solid = false;
	timer = 0;
	e->collection.addToCollection (getMonsterType());
}

void Monster::dummyAppear()
{
	state = MS_APPEAR;
	setAnim (teleport[1]);
	timer = 0;
}

void Monster::init(Resources *res)
{
	monkey[0] = res->getAnim ("monkey1");
	monkey[1] = res->getAnim ("monkey2");
	monkey[2] = res->getAnim ("monkey3");
	monkey[3] = res->getAnim ("monkey4");
	monkey[4] = res->getAnim ("monkey5");
	monkey[5] = res->getAnim ("monkey6");
	monkey[6] = res->getAnim ("monkey7");
	monkey[7] = res->getAnim ("monkey8");
	monkey[8] = res->getAnim ("monkey9");
	monkey[9] = res->getAnim ("monkey10");
	monkey[10] = res->getAnim ("monkey11");
	
	outline[0] = res->getAnim ("monkey1o");
	outline[1] = res->getAnim ("monkey2o");
	outline[2] = res->getAnim ("monkey3o");
	outline[3] = res->getAnim ("monkey4o");
	outline[4] = res->getAnim ("monkey5o");
	outline[5] = res->getAnim ("monkey6o");
	outline[6] = res->getAnim ("monkey7o");
	outline[7] = res->getAnim ("monkey8o");
	outline[8] = res->getAnim ("monkey9o");
	outline[9] = res->getAnim ("monkey10o");
	outline[10] = res->getAnim ("monkey11o");
	
	teleport[0] = res->getAnim ("monkeyskull");
	teleport[1] = res->getAnim ("monkeyskullreverse");

}
