#include <assert.h>
#include "player.h"
#include "resources.h"
#include "anim.h"
#include "engine.h"
#include "main.h"
#include "colors.h"
#include "monster.h"

Anim *Player::walk[2];
Anim *Bullet::bulletSprite;
Anim *PowerBonus::sprite;
Anim *Bonus::sprite;
Anim *Tire::sprite;

Bullet::Bullet(BITMAP *_field, int _dir, Player *_parent) : Object (_field)
{
	const int speed_factor = 6;
	visible = true;
	solid = true;
 	setAnim (bulletSprite);
	dx = (fix)(dir_mult[0][_dir] * speed_factor);
	dy = (fix)(dir_mult[1][_dir] * speed_factor);
	setDir (_dir);
	parent = _parent;
}

void Bullet::init(Resources *res)
{
	bulletSprite = res->getAnim("bullet");
}

void Bullet::update()
{
	try_move (dx, dy);
	
	Object::update();
}


void Bullet::handleCollission(Object *o)
{
	if (o->getType() == OT_MONSTER) // monster
	{
		kill();
	}
}
	
Player::Player(BITMAP *_field, PlayerState *_ps, int _playerType) : Object (_field)
{
	ps = _ps;
	transportCounter = 0;
	visible = true;
	hittimer = 0;
	attacktimer = 0;
	smoketimer = 0;
	button = NULL;
	playerType = _playerType;
	solid = true;
	
	setAnim (walk[playerType]);	
}

void Player::init(Resources *res)
{
	walk[0] = res->getAnim("fole");
	walk[1] = res->getAnim("raul");
}

void Player::update()
{
	if (!button) button = e->getButton(playerType);
	assert (button);
	assert (parent);
	
	if (ps->life > 0)
	{
		smoketimer++;
		if (smoketimer >= SMOKE_DELAY)
		{
			smoketimer = 0;
			e->smoke.add(getx(), gety());
		}
		
		if (octettimer > 0)
		{
			octettimer--;
		}
			
		if (hittimer > 0)
		{
			hittimer--;
			visible = !(hittimer & 1);
		}
		
		bool freshAnim = false;
			
		// attacking
		if (attacktimer > 0)
		{
			attacktimer --;
		}
		else
		{
			if (button[btnAction].getState())
			{
				attacktimer = e->ls.msecWeapon;
	// 			parent->getParent()->playSample (shoot);
	
				// generate bullet...
				Bullet *bullet = new Bullet(field, getDir(), this);
				parent->add (bullet);
				bullet->setLocation (getx() + 8, gety());
			}
		}
		
		fix dx = fix(0), dy = fix(0);
	
		const fix speedFactor = fix(2);
		// walking
		if (button[btnLeft].getState())
		{
			dx = -speedFactor;
			if (getDir() != LEFT) freshAnim = true;
			setDir(LEFT);
			isWalking = true;
		} else if (button[btnRight].getState())
		{
			dx = speedFactor;
			if (getDir() != RIGHT) freshAnim = true;
			setDir(RIGHT);
			isWalking = true;
		} else if (button[btnUp].getState())
		{
			dy = -speedFactor;
			if (getDir() != UP) freshAnim = true;
			setDir(UP);
			isWalking = true;
		} else if (button[btnDown].getState())
		{
			dy = speedFactor;
			if (getDir() != DOWN) freshAnim = true;
			setDir(DOWN);
			isWalking = true;
		} else
		{
			isWalking = false;
		}
		
		if (freshAnim)
		{
			setAnim (walk[playerType]);
		}
		
		try_move (dx, dy);
		
		int cx = getx() + getw() / 2;
		int cy = gety() + geth() / 2;
		circlefill (field, cx-8, cy-8, 10, MAGIC_PINK);
		circlefill (field, cx-8, cy+8, 10, MAGIC_PINK);
		circlefill (field, cx+8, cy-8, 10, MAGIC_PINK);
		circlefill (field, cx+8, cy+8, 10, MAGIC_PINK);
		
		// handle animation
		Object::update();
	}
}

void Player::die()
{
	ps->life--;
	if (ps->life == 0)
	{
		solid = false;
		visible = false;
	}
	else
	{
		hittimer = 20;
		setDefaultLocation();
	}
}

void Player::handleCollission (Object *o)
{
	if (ps->life == 0) return;
	switch (o->getType())
	{
	case OT_MONSTER:	
		if (e->getPowerMode()) // if in power mode
		{
			// monster will get collected
			ps->score += SCORE_MONSTER;
		}
		else
		{
			if (hittimer == 0) die();
		}
		break;
	case OT_POWER_BONUS:
			e->setPowerMode();
		break;
	case OT_TIRE:
		//TODO
		break;
	case OT_BONUS:
		ps->score += SCORE_BONUS; 
		if (octettimer > 0)
		{
			octetcount++;
			if (octetcount == 8)
			{
				ps->score += SCORE_OCTET;
			}
		}
		else
		{
			octettimer = OCTET_TIME;
			octetcount = 1;
		}
		break;
	case OT_BULLET: case OT_PLAYER:
		break; // do nothing
	}
}

PlayerState::PlayerState()
{
	life = 3;
	score = 0;
}

Bonus::Bonus(BITMAP *_field, int x, int y) : Object (_field)
{
	visible = true;
	solid = true;
 	setAnim (sprite);
	setDir (0);	
	setLocation ((fix)x * 32, (fix)y * 32);
}

void Player::setDefaultLocation()
{
	setLocation ((fix)(32 * ps->defaultx), (fix)(32 * ps->defaulty));
}

void Bonus::handleCollission(Object *o)
{
	if (o->getType() == OT_PLAYER)
	{
		kill();
	}
}

void Bonus::init(Resources *res)
{
	sprite = res->getAnim ("bonus1");
}

PowerBonus::PowerBonus(BITMAP *_field, int x, int y) : Bonus (_field, x, y)
{
	visible = true;
	solid = true;
 	setAnim (sprite);
	setDir (0);
}

void PowerBonus::init(Resources *res)
{
	sprite = res->getAnim ("bonus2");
}

Tire::Tire(BITMAP *_field, int x, int y) : Object (_field)
{
	visible = true;
	solid = true;
 	setAnim (sprite);
	setDir (0);
	dx = 0;
	dy = 0;
	setLocation ((fix)x * 32, (fix)y * 32);
}

void Tire::update()
{
	// fall if not on top of solid ground
	Object::update();
}

void Tire::handleCollission(Object *o)
{
}

void Tire::init(Resources *res)
{
	sprite = res->getAnim ("bonus3");
}
