#include <assert.h>
#include "resources.h"
#include "anim.h"
#include "dom.h"
#include "colors.h"

using namespace std;
using namespace xdom;

#ifndef PI
#define PI 3.1415926
#endif

// adds all filenames that match pattern to the vector
void glob(const char *pattern, vector<string> &result)
{
	char buf[1024];
	
	al_ffblk info;
	
	int temp = al_findfirst (pattern, &info, FA_RDONLY | FA_ARCH);
	
	while (temp == 0)
	{
		replace_filename(buf, pattern, info.name, sizeof(buf));
		result.push_back (std::string(buf));
		temp = al_findnext(&info);
	}
			
	al_findclose(&info);
}

void Resources::makeOutlineAnim (const string &srcName, const string &destName)
{
	Anim *src = getAnim (srcName);	
	Anim *a = new Anim();
	for (int dir = 0; dir < 4; ++dir)
	{
		Frame *srcFrame = src->getFrame (0, dir);		
		RLE_SPRITE * s = srcFrame->rle;
		BITMAP *temp = create_bitmap (s->w, s->h);	
		clear_to_color (temp, MAGIC_PINK);
		set_trans_blender (0, 0, 255, 255);
		draw_lit_rle_sprite (temp, s, 0, 0, 255);
		RLE_SPRITE *temp_rle = get_rle_sprite (temp);	
		destroy_bitmap (temp);
		generated.push_back (temp_rle);
		Frame f;
		f.rle = temp_rle;
		f.time = 0;
		f.w = srcFrame->w;
		f.h = srcFrame->h;
		f.hotx = srcFrame->hotx;
		f.hoty = srcFrame->hoty;
		a->add (f, dir);
		a->totalLength[dir] = 0;
	}
	a->totalDirs = src->totalDirs;
	animlist.insert (pair<string, Anim*>(destName, a));	
}

void Resources::makeHitAnim (const string &id, const string &id2)
{
	Anim *src = getAnim (id);
	Anim *a = new Anim();		
	for (int dir = 0; dir < 4; ++dir)
	{
		Frame *srcFrame = src->getFrame (0, dir);
		RLE_SPRITE * s = srcFrame->rle;
		BITMAP *temp = create_bitmap (s->w, s->h);
		clear_to_color (temp, MAGIC_PINK);
		draw_rle_sprite (temp, s, 0, 0);
		set_trans_blender (255, 255, 255, 255);
		draw_lit_rle_sprite (temp, s, 0, 0, 255);
		RLE_SPRITE *temp_rle = get_rle_sprite (temp);
		destroy_bitmap (temp);
		generated.push_back (temp_rle);
		Frame f;
		f.rle = temp_rle;
		f.time = 100;
		f.w = srcFrame->w;
		f.h = srcFrame->h;
		f.hotx = srcFrame->hotx;
		f.hoty = srcFrame->hoty;
		Frame f2;
		f2.rle = s;
		f2.time = 100;	
		f.w = srcFrame->w;
		f.h = srcFrame->h;
		f.hotx = srcFrame->hotx;
		f.hoty = srcFrame->hoty;
		a->add (f, dir);
		a->add (f2, dir);
	}
	a->totalLength[0] = 200;
	a->totalLength[1] = 200;
	a->totalLength[2] = 200;
	a->totalLength[3] = 200;
	a->totalDirs = src->totalDirs;
	animlist.insert (pair<string, Anim*>(id2, a));	
}

int Resources::init()
// returns 1 on succes, 0 on failure
{
	{
		vector <string> modfilenames;
		glob ("data/*.mod", modfilenames);
		for (vector<string>::iterator i = modfilenames.begin();
			i != modfilenames.end(); ++i)
		{
			unsigned int start = i->find_last_of ("\\/");
			if (start == string::npos) start = 0;
			else start += 1;
				
			unsigned int end = i->rfind (".");
			if (end == string::npos) end = i->size();
			else end -= 1;
			
			DUH *temp;
			if (!(temp = dumb_load_mod (i->c_str())))
			{
				return 0;
			}
			else
			{
				string name = i->substr(start, end-start+1);
				duhlist.insert (pair<string, DUH*>(name, temp));
			}
		}
	}

	{
		// load all .dat files in data directory
		vector <string> datafilenames;
		glob ("data/*.dat", datafilenames);
		for (vector<string>::iterator i = datafilenames.begin();
			i != datafilenames.end(); ++i)
		{
			DATAFILE *temp;
			if (!(temp = load_datafile (i->c_str())))
			{
				return 0;
			} else {
					data.push_back (temp);
			}
		}
	}
	vector <string> bmpfilenames;	
	glob ("data/*.bmp", bmpfilenames);
	
	for (vector<string>::iterator i = bmpfilenames.begin();
		i != bmpfilenames.end(); ++i)
	{
		unsigned int start = i->find_last_of ("\\/");
		if (start == string::npos) start = 0;
		else start += 1;
			
		unsigned int end = i->rfind (".");
		if (end == string::npos) end = i->size();
		else end -= 1;
			
		BITMAP *bmp;
		bmp = load_bitmap (i->c_str(), NULL);
		if (!bmp)
		{
			return 0;
		}
		else
		{
			string name = i->substr(start, end-start+1);
			sprites.insert (pair<string, RLE_SPRITE*>(name, get_rle_sprite(bmp)));
			destroy_bitmap (bmp);
		}
	}
		
	// load xml data
	
	vector <string> xmlfilenames;	
	glob ("data/*.xml", xmlfilenames);

	for (vector <string>::iterator i = xmlfilenames.begin(); i != xmlfilenames.end(); ++i)
	{
		DomNode node = xmlParseFile(*i);
		
		vector<DomNode>::iterator n;
		for (n = node.children.begin(); n != node.children.end(); ++n)
		{
			if (n->name == "anim")
			{
				Anim::loadFromXml (&(*n), this, animlist);
			}
			else
			{
				assert (false); // unknown tag
			}
		}
	}
	
	makeOutlineAnim ("monkey1", "monkey1o");
	makeOutlineAnim ("monkey2", "monkey2o");
	makeOutlineAnim ("monkey3", "monkey3o");
	makeOutlineAnim ("monkey4", "monkey4o");
	makeOutlineAnim ("monkey5", "monkey5o");
	makeOutlineAnim ("monkey6", "monkey6o");
	makeOutlineAnim ("monkey7", "monkey7o");
	makeOutlineAnim ("monkey8", "monkey8o");
	makeOutlineAnim ("monkey9", "monkey9o");
	makeOutlineAnim ("monkey10", "monkey10o");
	makeOutlineAnim ("monkey11", "monkey11o");
	return 1;
}

Resources::~Resources()
{
	{
		map<string, Anim*>::iterator i;
		for (i = animlist.begin(); i != animlist.end(); i++)
		{
			delete (i->second);
		}
	}
}

BITMAP *Resources::getBitmap (string id)
{
	vector<DATAFILE*>::iterator i;
	DATAFILE *temp = NULL;
	for (i = data.begin(); i != data.end() && temp == NULL; ++i)
	{
			assert (*i);
			temp = find_datafile_object (*i, id.c_str());				
			// check if object is of the right type
			if (temp && temp->type != DAT_BITMAP)
					temp = NULL;
	}
	if (temp)
		return (BITMAP*)temp->dat;
	else
	{
		allegro_message ("Couldn't find bitmap '%s'", id.c_str());			
		exit(1);
	}
	return NULL;
}

FONT *Resources::getFont (string id)
{
	vector<DATAFILE*>::iterator i;
	DATAFILE *temp = NULL;
	for (i = data.begin(); i != data.end() && temp == NULL; ++i)
	{
			assert (*i);
			temp = find_datafile_object (*i, id.c_str());				
			// check if object is of the right type
			if (temp && temp->type != DAT_FONT)
					temp = NULL;
	}
	if (temp)
		return (FONT*)temp->dat;
	else
	{
		allegro_message ("Couldn't find FONT '%s'", id.c_str());			
		exit(1);
	}
	return NULL;	
}

RLE_SPRITE *Resources::getRle (string id)
{
	vector<DATAFILE*>::iterator i;
	DATAFILE *temp = NULL;
	for (i = data.begin(); i != data.end() && temp == NULL; ++i)
	{
			assert (*i);
			temp = find_datafile_object (*i, id.c_str());				
			// check if object is of the right type
			if (temp && temp->type != DAT_RLE_SPRITE)
					temp = NULL;
	}
	
	if (temp)
			return (RLE_SPRITE*)temp->dat;
	else
	{
		if (sprites.find (id) == sprites.end())
		{
			allegro_message ("Couldn't find RLE sprite '%s'", id.c_str());			
			exit(1);
		}
		
		return sprites[id];
	}
	
	assert (false);
	return NULL;
}

SAMPLE *Resources::getSample (string id)
{
	vector<DATAFILE*>::iterator i;
	DATAFILE *temp = NULL;
	for (i = data.begin(); i != data.end() && temp == NULL; ++i)
	{
			assert (*i);
			temp = find_datafile_object (*i, id.c_str());				
			// check if object is of the right type
			if (temp && temp->type != DAT_SAMPLE)
					temp = NULL;
	}
	if (temp)
			return (SAMPLE*)temp->dat;
	else
	{
		allegro_message ("Couldn't find SAMPLE '%s'", id.c_str());			
		exit(1);
	}		
	return NULL;
}

Anim *Resources::getAnim (string id)
{
	if (animlist.find (id) == animlist.end())
	{
		allegro_message ("Couldn't find Animation '%s'", id.c_str());			
		exit(1);
	}
	else
	{
		return animlist[id];
	}
}

DUH *Resources::getDuh (string id)
{
    if (duhlist.find (id) == duhlist.end())
    {
		allegro_message ("Couldn't find DUH (Music) '%s'", id.c_str());	
		exit(1);
    }
    return duhlist[id];
}
